from . import BaseActor
from ltr.models.loss.target_regression import DenseIouPred
import torch
from ltr.data.processing_utils import iou_xyxy
import cv2
import matplotlib.pyplot as plt
import os

class ClsActor(BaseActor):
    """Actor for jointly training cls_18, cls_72 branch and regression branch, except for regression optimizer"""
    def __init__(self, net, objective, device, output_sz, loss_weight=None):
        super().__init__(net, objective, device)
        self.loss_weight = loss_weight
        self.output_sz = output_sz

    def box_cxcywh_to_xyxy(self, x):
        x_c, y_c, w, h = x.unbind(-1)
        b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
             (x_c + 0.5 * w), (y_c + 0.5 * h)]
        return torch.stack(b, dim=-1)

    def box_xywh_to_xyxy(self, x):
        x1, y1, w, h = x.unbind(-1)
        b = [x1, y1, x1 + w, y1 + h]
        return torch.stack(b, dim=-1)

    def __call__(self, data, vis=False):
        """
        args:
            data - The input data.
            gen_iou_map -
        returns:
            loss    - the training loss
            stats  -  dict containing detailed losses
        """
        # Run network
        # print(data['test_label'].shape)
        output = self.net(train_imgs=data['train_images'],
                          test_imgs=data['test_images'],
                          train_bb=data['train_anno'])

        ### Regression offline loss ###
        # print(output_coord.shape)
        # gt_box = data['test_anno'].view(-1, 4)
        # gt_center = (gt_box[:, :2] + 0.5 * gt_box[:, 2:]) / self.output_sz  # nb,

        # l1_loss = self.objective['l1'](output['center'], gt_center) * self.loss_weight['l1']
        cls_loss = self.objective['test_clf_s'](output['scores_s'], data['test_label_s'].float().detach())

        # clf_losses = {}
        # clf_final_loss = {}
        # clf_init_loss = {}
        # clf_iter_loss = {}
        # for k, scores in output.items():
        #     if k == 'scores_s':
        #         continue
        #     ### Classification-18 losses ###
        #     clf_losses[k] = [self.objective['test_clf_18'](s, data['test_label'], data['test_anno']) for s in scores]
        #
        #     # Loss of the final filter
        #     clf_final_loss[k] = clf_losses[k][-1]
        #     clf_final_loss[k] = self.loss_weight['test_clf_18'] * clf_final_loss[k]
        #
        #     # Loss for the initial filter iteration
        #     clf_init_loss[k] = 0
        #     if 'test_init_clf_18' in self.loss_weight.keys():
        #         clf_init_loss[k] = self.loss_weight['test_init_clf_18'] * clf_losses[k][0]
        #
        #     # Loss for the intermediate filter iterations
        #     clf_iter_loss[k] = 0
        #     if 'test_iter_clf_18' in self.loss_weight.keys():
        #         test_iter_weights_18 = self.loss_weight['test_iter_clf_18']
        #         if isinstance(test_iter_weights_18, list):
        #             clf_iter_loss[k] = sum(
        #                 [a * b for a, b in zip(test_iter_weights_18, clf_losses[k][1:-1])])
        #         else:
        #             clf_iter_loss[k] = (test_iter_weights_18 / (len(clf_losses[k]) - 2)) * sum(
        #                 clf_losses[k][1:-1])

        vis_results = {}
        if vis:
            vis_score_maps = self.tensors_gray2rgb(output['scores_s'][:, :1, ...])
            vis_results['score_maps'] = vis_score_maps
            gt_score_maps = self.tensors_gray2rgb(data['test_label_s'][:, :1, ...])
            vis_results['gt_score_maps'] = gt_score_maps

        # Total loss
        # loss = clf_final_loss['scores_3'] + clf_iter_loss['scores_3'] + clf_init_loss['scores_3'] + \
        #     cls_loss * self.loss_weight['test_clf_s'] + \
        #     clf_final_loss['scores_1'] + clf_iter_loss['scores_1'] + clf_init_loss['scores_1'] + \
        #     clf_final_loss['scores_2'] + clf_iter_loss['scores_2'] + clf_init_loss['scores_2']
        loss = cls_loss * self.loss_weight['test_clf_s']


        # Log stats
        stats = {'Loss/total': loss.item(),
                 'Loss/cls_s': cls_loss.item(),
                 # 'Loss/cls_final_loss_1': clf_final_loss['scores_1'].item(),
                 # 'Loss/cls_init_loss_1': clf_init_loss['scores_1'].item(),
                 # 'Loss/cls_iter_loss_1': clf_iter_loss['scores_1'].item(),
                 # 'Loss/cls_final_loss_2': clf_final_loss['scores_2'].item(),
                 # 'Loss/cls_init_loss_2': clf_init_loss['scores_2'].item(),
                 # 'Loss/cls_iter_loss_2': clf_iter_loss['scores_2'].item(),
                 # 'Loss/cls_final_loss_3': clf_final_loss['scores_3'].item(),
                 # 'Loss/cls_init_loss_3': clf_init_loss['scores_3'].item(),
                 # 'Loss/cls_iter_loss_3': clf_iter_loss['scores_3'].item(),
                }


        return loss, stats, vis_results

    def tensors_gray2rgb(self, tensors):
        """Convert to rgb images [tenosrs： (num, 1, h, w)].
           :return tensors： (num, 3, h, w)
        """

        tmp = '/home/cyt/project/FCOT/temp.png'
        out = []
        for tensor in tensors:
            image = tensor.clone().squeeze().cpu().detach().numpy()  # [h, w]
            plt.imsave(tmp, image)
            rgb = plt.imread(tmp)[..., 0:3]      # [h, w, 3]
            rgb = torch.from_numpy(rgb).permute(2, 0, 1)
            out.append(rgb)
            # print(rgb.shape)
        out = torch.stack(out, dim=0).float().to(tensors.device)

        return out

    def convert_img_tensor_to_vis_format(self, img_tensors):
        """
        :param img_tensors: n, b, c, h, w
        :return:
        """
        _mean = torch.Tensor([0.485, 0.456, 0.406]).view(-1, 1, 1)
        _std = torch.Tensor([0.229, 0.224, 0.225]).view(-1, 1, 1)
        img_tensor = img_tensors[0, 0].clone().cpu() * _std + _mean
        img_tensor *= 255.0
        img = img_tensor.permute(1, 2, 0).numpy()  # h, w, 3
        return img

    def draw_image(self, img, bboxs):
        """Display a 2D tensor.
        """
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        colors = [(0,0,255),
                  (0,255,0),
                  (255,0,0),
                  (255,255,0),
                  (255,0,255),
                  (0,255,255),
                  (0,0,125),
                  (0,125,0),
                  (125,0,0),
                  (125,125,125)]
        i = 0
        for bbox in bboxs:
            img = cv2.rectangle(img, (bbox[0], bbox[1]), (bbox[2], bbox[3]), colors[i], 2)
            i += 1
        # draw_rec = cv2.circle(draw_rec, (center[0], center[1]), 3, (0, 0, 255), -1)
        # draw_rec = cv2.circle(draw_rec, (center2[1], center2[0]), 3, (0, 255, 0), -1)

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img_tensor = torch.from_numpy(img).float().permute(2, 0, 1).unsqueeze(0)
        return img_tensor

