from . import BaseActor
from ltr.models.loss.target_regression import DenseIouPred
import torch
import cv2

class PropsSeActor(BaseActor):
    """Actor for jointly training cls_18, cls_72 branch and regression branch, except for regression optimizer"""
    def __init__(self, net, propsals_generator, objective, device, output_sz, loss_weight=None):
        super().__init__(net, objective, device)
        self.loss_weight = loss_weight
        self.output_sz = output_sz
        self.propsals_generator = propsals_generator

    def box_xywh_to_xyxy(self, x):
        x1, y1, w, h = x.unbind(-1)
        b = [x1, y1, x1 + w, y1 + h]
        return torch.stack(b, dim=-1)

    def box_cxcywh_to_xyxy(self, x):
        x_c, y_c, w, h = x.unbind(-1)
        b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
             (x_c + 0.5 * w), (y_c + 0.5 * h)]
        return torch.stack(b, dim=-1)

    def __call__(self, data, vis=False):
        """
        args:
            data - The input data.
            gen_iou_map -
        returns:
            loss    - the training loss
            stats  -  dict containing detailed losses
        """
        # generate proposals using other tracker
        re = self.propsals_generator.generate_propsals(train_imgs=data['train_images'],
                                                       test_imgs=data['test_images'],
                                                       train_bb=data['train_anno'],
                                                       test_bb=data['test_anno'])
        #目前需要保证二者的num_images相等
        assert data['test_anno'].shape == data['train_anno'].shape
        n, b, _ = data['test_anno'].size()
        # Run network
        test_pos_props_s, test_neg_props_s = self.net(train_imgs=data['train_images'],
                                                       test_imgs=data['test_images'],
                                                       proposals=re)

        ### bbox loss loss ###
        # print("pos_s: {}".format(test_pos_props_s))
        # print("neg_s: {}".format(test_neg_props_s))

        pos_loss = self.objective['score'](test_pos_props_s, re['test_pos_props_label'].detach()) * self.loss_weight['score']
        neg_loss = self.objective['score'](test_neg_props_s, re['test_neg_props_label'].detach()) * self.loss_weight['score']

        draw_img = None
        vis_results = {}
        if vis:
            imgs = self.convert_img_tensor_to_vis_format(data['test_images']) # n, h, w, 3
            output_bb = self.box_xywh_to_xyxy(re['test_pos_props'])[:, 0, :].clone()
            bbs = output_bb.cpu().int().numpy()  # n, 4
            draw_img_test_pos = self.draw_images(imgs, bbs)  # n, c, h, w

            imgs = self.convert_img_tensor_to_vis_format(data['test_images'].repeat(2, 1, 1, 1, 1))
            output_bb = self.box_xywh_to_xyxy(re['test_neg_props'])[:, 0, :].clone()
            bbs = output_bb.cpu().int().numpy()  # 2n, 4
            draw_img_test_neg = self.draw_images(imgs, bbs)  # 1, c, h, w

            imgs = self.convert_img_tensor_to_vis_format(data['train_images'])
            output_bb = self.box_xywh_to_xyxy(re['train_gt_props'])[:, 0, :].clone()
            bbs = output_bb.cpu().int().numpy()  # n, 4
            draw_img_train_gt = self.draw_images(imgs, bbs)  # 1, c, h, w

            vis_results['train_gt'] = draw_img_train_gt
            vis_results['test_pos'] = draw_img_test_pos
            vis_results['test_neg'] = draw_img_test_neg

        # Total loss
        loss =  neg_loss + pos_loss

        # Log stats
        stats = {'Loss/total': loss.item(),
                 'Loss/pos_props': pos_loss.item(),
                 'Loss/neg_props': neg_loss.item()
                }

        return loss, stats, vis_results

    def convert_img_tensor_to_vis_format(self, img_tensors):
        """
        :param img_tensors: n, b, c, h, w
        :return: n, h, w, 3
        """
        _mean = torch.Tensor([0.485, 0.456, 0.406]).view(1, -1, 1, 1)
        _std = torch.Tensor([0.229, 0.224, 0.225]).view(1, -1, 1, 1)
        img_tensor = img_tensors[:, 0].clone().cpu() * _std + _mean
        img_tensor *= 255.0
        img = img_tensor.permute(0, 2, 3, 1).numpy()  # n, h, w, 3
        return img

    def draw_images(self, imgs, bboxs):
        """Display a 4D tensor.
        """
        colors = [(0,0,255),
                  (0,255,0),
                  (255,0,0),
                  (255,255,0),
                  (255,0,255),
                  (0,255,255),
                  (0,0,125),
                  (0,125,0),
                  (125,0,0),
                  (125,125,125)]
        imgs_tensor = []
        i = 0
        for bbox in bboxs:
            img = cv2.cvtColor(imgs[i], cv2.COLOR_RGB2BGR)
            img = cv2.rectangle(img, (bbox[0], bbox[1]), (bbox[2], bbox[3]), colors[i], 2)
            i += 1
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            img_tensor = torch.from_numpy(img).float().permute(2, 0, 1)
            imgs_tensor.append(img_tensor)
        # draw_rec = cv2.circle(draw_rec, (center[0], center[1]), 3, (0, 0, 255), -1)
        # draw_rec = cv2.circle(draw_rec, (center2[1], center2[0]), 3, (0, 255, 0), -1)
        imgs_tensor = torch.stack(imgs_tensor, dim=0)

        return imgs_tensor

