from torch.nn.parallel import DistributedDataParallel as DDP
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict
import random

from ltr.data.processing_utils import iou, rand_uniform
from ltr.models.target_classifier.initializer import FilterPool
from pytracking.utils.loading import load_network

class Propsals_generator(object):
    def __init__(self, props_generator_dir, settings, props_sz=4):
        self.props_generator_net = self.get_network(props_generator_dir)
        self.feat_pool_1 = FilterPool(filter_size=props_sz, feature_stride=4, pool_square=False)  # layer1 feat (nb, 256, 72, 72)
        self.feat_pool_2 = FilterPool(filter_size=props_sz, feature_stride=8, pool_square=False)  # layer2 feat (nb, 512, 36, 36)
        self.feat_pool_3 = FilterPool(filter_size=props_sz, feature_stride=16, pool_square=False)  # layer3 feat (nb, 1024, 18, 18)
        self.settings = settings

    def get_network(self, checkpoint_dir):
        network = load_network(checkpoint_dir)
        network.cuda()
        network.eval()
        if self.settings.local_rank != -1:
            network = torch.nn.SyncBatchNorm.convert_sync_batchnorm(network)
            network = DDP(network, device_ids=[self.settings.local_rank], find_unused_parameters=True)
        return network

    def generate_propsals(self, train_imgs, test_imgs, train_bb, test_bb):
        """
        当前版本1的实现：对于每个batch中的N（N=3）张images, 训练集中只采样正样本，其中第一帧进行类似于测试过程中取gt_box，剩余N-1帧对gt框添加一定的噪声（iou>0.85）来模拟online过程中的正样本；
        对于测试集中的每一帧，生成1个正样本（1个iou>0.85的框），生成2个负样本（1个易混淆目标+1个iou<0.3的框）
        TODO： 其中对于训练proposals添加两种embedding，来标记online和第一帧的权重
        :param train_imgs: (n, b, c, h, w)
        :param test_imgs: (n, b, c, h, w)
        :param train_bb: (n, b, 4)
        :param test_bb: (n, b, 4)
        :return: train_gt_proposals (n, b, 4), train_gt_labels (n, b, 1),
                 test_pos_proposals (n, b, 4), test_pos_labels (n, b, 1),
                 test_neg_proposals (2n, b, 4),, test_neg_labels (2n, b, 1),
        """
        output = self.props_generator_net.forward(train_imgs, test_imgs, train_bb)
        scores_72 = output['target_scores_72']
        offset_maps = output['offset_maps']
        train_backbone_feat = output['train_feat']
        test_backbone_feat = output['test_feat']

        # get test negative proposals and labels
        n, b, _ = test_bb.size()
        test_neg_props1 = self.get_neg_prop(scores_72[-1], offset_maps, test_bb) # n, b, 4
        test_neg_props2 = [self.perturb_box(bb.clone(), max_iou=0.3, min_iou=0.0, sigma_factor=0.5) for bb in test_bb.view(-1, 4)]
        test_neg_props2 = torch.stack(test_neg_props2, dim=0).view(n, b, 4) # n, b, 4
        test_neg_props = torch.cat([test_neg_props1, test_neg_props2], dim=0) # 2n, b, 4
        test_neg_props_label = torch.zeros(2*n, b).to(test_bb.device)
        test_neg_backbone_feat = {k:feat.view(n, b, *feat.shape[-3:]).repeat(2, 1, 1, 1, 1).view(-1, *feat.shape[-3:]) for k,feat in test_backbone_feat.clone().items()}
        test_neg_props_feat = self.generate_props_feat(test_neg_backbone_feat, test_neg_props)  # layer1: (2n*b, c, 4, 4) etc.

        # get test positive proposals
        test_pos_props = [self.perturb_box(bb.clone(), max_iou=1.0, min_iou=0.85, sigma_factor=0.05) for bb in test_bb.view(-1, 4)]
        test_pos_props = torch.stack(test_pos_props, dim=0).view(n, b, 4) # n, b, 4
        test_pos_props_label = torch.ones(n, b).to(test_bb.device)
        test_pos_props_feat = self.generate_props_feat(test_backbone_feat, test_pos_props)  # layer1: (n*b, c, 4, 4) etc.

        # get train gt proposals
        n, b, _ = train_bb.size()
        train_gt_props0 = train_bb[:1].clone()  # 1, b, 4
        train_gt_props_online = [self.perturb_box(bb.clone(), max_iou=1.0, min_iou=0.85, sigma_factor=0.05) for bb in train_bb[1:].view(-1, 4)]
        train_gt_props_online = torch.stack(train_gt_props_online, dim=0).view(-1, b, 4) # n-1, b, 4
        train_gt_props = torch.cat([train_gt_props0, train_gt_props_online], dim=0)  # n, b, 4
        # train_gt_props_label = torch.ones(n, b, 1).to(train_bb.device)
        train_gt_props_feat = self.generate_props_feat(train_backbone_feat, train_gt_props)  # layer1: (n*b, c, 4, 4) etc.

        re = {}
        re['test_neg_props'] = test_neg_props
        re['test_neg_props_label'] = test_neg_props_label
        re['test_neg_props_feat'] = test_neg_props_feat

        re['test_pos_props'] = test_pos_props
        re['test_pos_props_label'] = test_pos_props_label
        re['test_pos_props_feat'] = test_pos_props_feat

        re['train_gt_props'] = train_gt_props
        # re['train_gt_props_label'] = train_gt_props_label
        re['train_gt_props_feat'] = train_gt_props_feat

        return re


    def generate_props_feat(self, feat, bb, mode='all'):
        """

        :param feat: layer1/2/3: [nb, c, h, w]
        :param bb: [n, b, 4]
        :return:
        """
        output_feats = {}
        bb_pool = bb.clone()

        # print("bb: {}".format(bb))
        if mode == 'all':
            output_feats['layer1'] = self.feat_pool_1(feat['layer1'], bb_pool)  # n, b, 256, 4, 4
            output_feats['layer2'] = self.feat_pool_2(feat['layer2'], bb_pool)  # n, b, 512, 4, 4
            output_feats['layer3'] = self.feat_pool_3(feat['layer3'], bb_pool)  # n, b, 1024, 4, 4
        else:
            raise Exception("other modes are not support now.")

        return output_feats


    def get_neg_prop(self, cls_hm, offset_maps, bb, neg_iou_thre=0.15, scale=1/4):
        '''
            cls_hm: num, b, 72, 72
            offset_maps: num, b, 4(l,r,t,b), 72, 72
            bb: num, b, 4
        '''
        N, B = cls_hm.shape[:2]

        offset_maps = offset_maps.detach()
        cls_hm = cls_hm.detach()

        gt_bb = bb.clone() * scale
        cls_hm_72 = F.interpolate(cls_hm.view(-1,*cls_hm.shape[-2:]).unsqueeze(1), size=72, mode='bilinear', align_corners=True) # n, b, 1, 72*72
        x_coors = torch.arange(72).view(1,1,1,-1).repeat(N,B,72,1).float().to(cls_hm_72.device) + 0.5
        y_coors = torch.arange(72).view(1,1,-1,1).repeat(N,B,1,72).float().to(cls_hm_72.device)  + 0.5
        wl = x_coors.unsqueeze(2) - offset_maps[:,:,0:1,:,:] # x0
        wr = x_coors.unsqueeze(2) + offset_maps[:,:,1:2,:,:] # x1
        ht = y_coors.unsqueeze(2) - offset_maps[:,:,2:3,:,:] # y0
        hb = y_coors.unsqueeze(2) + offset_maps[:,:,3:,:,:] # y1
        pred_x0y0x1y1 = torch.cat([wl, ht, wr, hb], dim=2).view(N,B,4,-1).view(-1,4,72*72) # nb, 4, 72*72
        cls_hm_72 = cls_hm_72.view(N,B,1,-1).view(-1,1,72*72).repeat(1,4,1) # nb, 4, 72*72
        gt_x0y0x1y1 = gt_bb.clone()
        gt_x0y0x1y1[:,:,2:] = gt_x0y0x1y1[:,:,2:] + gt_x0y0x1y1[:,:,:2] # nb, 4, 1
        gt_x0y0x1y1 = gt_x0y0x1y1.view(-1, 4).unsqueeze(-1)
        ious =  iou(gt_x0y0x1y1, pred_x0y0x1y1).unsqueeze(1).repeat(1,4,1) # nb, 4, 72*72
        # neg_inds = ious < neg_iou_thre

        # max1 = torch.max(cls_hm_72, dim=-1)[0][:,0] # (nb,)

        mask_inds = ious > neg_iou_thre
        cls_hm_72[mask_inds] = -1e6

        # max2 = torch.max(cls_hm_72, dim=-1)[0][:,0]

        max_neg_ind = torch.max(cls_hm_72, dim=-1)[1].unsqueeze(-1) # nb, 4, 1
        # print(cls_hm_72.shape)
        # print(max_neg_ind.shape)
        # print(pred_x0y0x1y1.shape)
        neg_selected = torch.gather(pred_x0y0x1y1, 2, max_neg_ind) # nb, 4, 1
        neg_selected = neg_selected.reshape(N, B, 4) / scale
        return neg_selected



    def perturb_box(self, box, min_iou=0.85, max_iou=1.0, sigma_factor=0.1):
        """ Perturb the input box by adding gaussian noise to the co-ordinates

         args:
            box - input box [n, b, 4]
            min_iou - minimum IoU overlap between input box and the perturbed box
            sigma_factor - amount of perturbation, relative to the box size. Can be either a single element, or a list of
                            sigma_factors, in which case one of them will be uniformly sampled. Further, each of the
                            sigma_factor element can be either a float, or a tensor
                            of shape (4,) specifying the sigma_factor per co-ordinate

        returns:
            torch.Tensor - the perturbed box
        """

        if isinstance(sigma_factor, list):
            # If list, sample one sigma_factor as current sigma factor
            c_sigma_factor = random.choice(sigma_factor)
        else:
            c_sigma_factor = sigma_factor

        if not isinstance(c_sigma_factor, torch.Tensor):
            c_sigma_factor = c_sigma_factor * torch.ones(4)

        perturb_factor = torch.sqrt(box[2] * box[3]) * c_sigma_factor

        # multiple tries to ensure that the perturbed box has iou > min_iou with the input box
        for i_ in range(1000):
            c_x = box[0] + 0.5 * box[2]
            c_y = box[1] + 0.5 * box[3]
            c_x_per = random.gauss(c_x, perturb_factor[0])
            c_y_per = random.gauss(c_y, perturb_factor[1])

            w_per = random.gauss(box[2], perturb_factor[2])
            h_per = random.gauss(box[3], perturb_factor[3])

            if w_per <= 1:
                w_per = box[2] * rand_uniform(0.15, 0.5)

            if h_per <= 1:
                h_per = box[3] * rand_uniform(0.15, 0.5)

            box_per = torch.Tensor([c_x_per - 0.5 * w_per, c_y_per - 0.5 * h_per, w_per, h_per]).round()

            if box_per[2] <= 1:
                box_per[2] = box[2] * rand_uniform(0.15, 0.5)

            if box_per[3] <= 1:
                box_per[3] = box[3] * rand_uniform(0.15, 0.5)

            box_iou = iou(box.view(1, 4), box_per.view(1, 4))
            # print("iou: {}".format(box_iou))

            # if there is sufficient overlap, return
            if box_iou >= min_iou and box_iou <= max_iou:
                return box_per, box_iou

        return None