from pytracking.tracker.base import BaseTracker
import torch
import torch.nn.functional as F
import torch.nn as nn
import math
import time
from pytracking import dcf, TensorList
from pytracking.features.preprocessing import numpy_to_torch
from pytracking.utils.plotting import show_tensor, plot_graph
from pytracking.features.preprocessing import sample_patch_multiscale, sample_patch_transformed, sample_patch
from pytracking.features import augmentation
import cv2
import random
import os
import matplotlib.pyplot as plt
import numpy as np
import torch.optim as optim
import ltr.models.loss as ltr_losses
from ltr.models.loss.target_regression import REGLoss
import ltr.data.processing_utils as prutils
import copy
from .trSeg_utils import rect_from_mask, mask_from_rect
from .trSeg_utils import TrSegModule

#seed = 1234
#torch.manual_seed(seed)
#torch.cuda.manual_seed(seed)
#torch.cuda.manual_seed_all(seed)  # if you are using multi-GPU.
#np.random.seed(seed)  # Numpy module.
#random.seed(seed)  # Python random module.
#torch.manual_seed(seed)

#torch.backends.cudnn.deterministic = True
#torch.backends.cudnn.benchmark = False
torch.backends.cudnn.enabled = True


class TrSeg(BaseTracker):

    def initialize_features(self):
        if not getattr(self, 'features_initialized', False):
            self.params.net.initialize()
        self.features_initialized = True

    def convert_bbox(self, bbox):
        """
        :param bbox: array [x0, y0, w, h]
        :return: tensors: center pos (yc, xc) and sz (h, w)
        """
        pos = torch.Tensor([bbox[1] + (bbox[3] - 1) / 2, bbox[0] + (bbox[2] - 1) / 2])  # y, x
        target_sz = torch.Tensor([bbox[3], bbox[2]])
        return pos, target_sz

    def initialize(self, image, info: dict) -> dict:
        # Initialize some stuff
        self.frame_num = 1
        if not hasattr(self.params, 'device'):
            self.params.device = 'cuda' if self.params.use_gpu else 'cpu'

        # Initialize network
        # self.initialize_features()

        # The DiMP network
        # self.net = self.params.net
        self.net = copy.deepcopy(self.params.net)
        # self.transCls_net = copy.deepcopy(self.params.transCls_net)

        # Time initialization
        tic = time.time()

        if 'init_box' in info.keys():
            mask = None
        elif 'init_mask' in info.keys():
            mask = info['init_mask']  # H W
            info['init_bbox'] = rect_from_mask(mask)  # x0, y0, w, h

        # Get target position and size
        state = info['init_bbox']  # x y w h
        self.pos, self.target_sz = self.convert_bbox(state)

        # Set sizes
        sz = self.params.image_sample_size  # 288
        self.feat_sz = int(sz / 4)
        self.img_sample_sz = torch.Tensor([sz, sz] if isinstance(sz, int) else sz)
        self.img_support_sz = self.img_sample_sz  # [288, 288]

        # Set search area
        search_area = torch.prod(self.target_sz * self.params.search_area_scale).item()
        self.target_scale = math.sqrt(search_area) / self.img_sample_sz.prod().sqrt()

        # Target size in base scale
        self.base_target_sz = self.target_sz / self.target_scale

        # initialize trseg module
        self.initialize_trseg(image, state)

        # Convert image
        im = numpy_to_torch(image)

        # Setup scale factors
        if not hasattr(self.params, 'scale_factors'):
            self.params.scale_factors = torch.ones(1)
        elif isinstance(self.params.scale_factors, (list, tuple)):
            self.params.scale_factors = torch.Tensor(self.params.scale_factors)

        # Setup scale bounds
        self.image_sz = torch.Tensor([im.shape[2], im.shape[3]])
        self.min_scale_factor = torch.max(10 / self.base_target_sz)
        self.max_scale_factor = torch.min(self.image_sz / self.base_target_sz)

        # self.params.transcls_mem_size = getattr(self.params, 'transcls_mem_size', 3)

        # Extract and transform sample
        init_backbone_feat = self.generate_init_samples(im)

        # Initialize classifiers and regressor
        self.init_classifier(init_backbone_feat)
        # self.init_transCls(init_transcls_backbone_feat)

        self.prev_pos = self.pos
        self.prev_box = torch.Tensor([state[1], state[0], state[1] + state[3], state[0] + state[2]])
        self.train_x_3 = None
        self.train_x_23 = None
        self.train_x_123 = None
        self.target_box = None
        self.s = None
        self.max_score = -100
        self.max_score_mask = -100

        self.seq_id = 0

        out = {'time': time.time() - tic}

        self.last_update_mask = self.frame_num
        return out

    def init_transCls(self, backbone_feat):
        with torch.no_grad():
            feat_transCls = backbone_feat['layer3'][:self.params.transcls_mem_size]  #
            # print("feat shape: {}".format(feat_transCls.shape))
            train_feat_s = self.transCls_net.proj3(feat_transCls.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
            target = train_feat_s.view(self.params.transcls_mem_size, *train_feat_s.shape[-3:])
            bb_pool = self.target_boxes[:self.params.transcls_mem_size].clone()
            target = self.transCls_net.target_pool(target, bb_pool)  # 3, C, Ht, Wt
            target = target.view(self.params.transcls_mem_size, -1, 4, 4)
            target = target.flatten(2).permute(2, 0, 1)  #
            target = self.transCls_net.transformer.encoder2(target)
            target = self.transCls_net.transformer.box_head1(target.permute(2, 1, 0)).permute(2, 1, 0)  # 1, 3, C  # 3 is the number of targets
            target = target.view(self.params.transcls_mem_size, 1, -1)
            self.transCls_target = target   # [3, 1, C]
            self.previous_replace_ind_transcls = 1
            self.train_x_transCls = None
            # search = test_feat_s.flatten(start_dim=2).permute(2, 0, 1)  # HWxNBxC
            # hs = self.transformer(search, target)  # HWxNBxC
            # score_s = self.score_head(hs).squeeze().permute(1, 0).view(n, b, *test_feat_s.shape[-2:])

    def initialize_trseg(self, frame1, bbox1):
        trseg_net_dir = getattr(self.params, 'trseg_net_dir', None)
        assert trseg_net_dir is not None
        self.init_state = bbox1
        mask_search_area_scale = getattr(self.params, 'mask_search_area_scale', 2.0)
        self.trseg = TrSegModule(trseg_net_dir=trseg_net_dir, search_factor=mask_search_area_scale, input_sz=256)
        self.trseg.initialize(frame1, np.array(bbox1))

    def track(self, image, vis_switch=False) -> dict:
        self.debug_info = {}

        self.frame_num += 1
        self.debug_info['frame_num'] = self.frame_num
        print("---------track frame-{}-----------".format(self.frame_num))

        # Convert image to tensor [1, 3, w, h]
        im = numpy_to_torch(image)
        ###
        # new_state = torch.cat((self.pos[[1, 0]] - (self.target_sz[[1, 0]] - 1) / 2, self.target_sz[[1, 0]]))  # x, y, w, h
        ###
        # ------- LOCALIZATION ------- #
        search_prev_pos = self.prev_pos.clone()
        backbone_feat, sample_coords = self.extract_backbone_features(im,
                                                                      self.prev_pos,
                                                                      self.target_scale * self.params.scale_factors,
                                                                      self.img_sample_sz)

        # Extract classification features
        with torch.no_grad():
            test_feat_3, test_feat_23, test_feat_123 = self.net.extract_3scales_feat(backbone_feat)
            test_feat_3 = self.net.classifier_3.extract_classification_feat(test_feat_3)
            test_feat_23 = self.net.classifier_2.extract_classification_feat(test_feat_23)
            test_feat_123 = self.net.classifier_1.extract_classification_feat(test_feat_123)
            # print(test_feat_3.shape, test_feat_23.shape, test_feat_123.shape)
            assert test_feat_3.shape == test_feat_23.shape and test_feat_3.shape == test_feat_123.shape

        # Location of sample
        sample_pos, sample_scales = self.get_sample_location(sample_coords)

        # Compute classification scores
        scores_raw_3 = self.classify_target_3(test_feat_3)
        scores_raw_23 = self.classify_target_23(test_feat_23)
        scores_raw_123 = self.classify_target_123(test_feat_123)
        # scores = scores_raw_3 * self.params.merge_rate_18 + scores_raw_23 * 0.2 + scores_raw_123 * 0.1
        # scores = scores_raw_3 + scores_raw_23 * 0.2 + scores_raw_123 * 0.1
        scores = scores_raw_3 + scores_raw_23 * 0.15 + scores_raw_123 * 0.05

        # locate the center pos
        translation_vec, scale_ind, s, flag, max_disp, max_disp2, max_score = \
            self.localize_target(scores, sample_scales, feature_sz=test_feat_3.size(-1))
        # translation_vec, scale_ind, s, flag, max_disp, max_disp2, max_score = \
        #     self.localize_target(scores_72, sample_scales, feature_sz=self.feat_sz)
        new_pos = sample_pos[scale_ind, :] + translation_vec

        # assert self.frame_num <= 50

        self.debug_info['flag'] = flag
        print("flag: {}".format(flag))
        # if self.frame_num == 100:
        #     assert 0

        # Update position and scale
        if flag != 'not_found':
            self.prev_pos = new_pos.clone()
            self.pos = self.prev_pos

        score_map = s[scale_ind, ...]

        # Compute stage-1 bounding box
        ###
        # new_state = torch.cat(
        #     (self.pos[[1, 0]] - (self.target_sz[[1, 0]] - 1) / 2, self.target_sz[[1, 0]]))  # x, y, w, h
        new_state = torch.cat((self.pos[[1, 0]] - (self.target_sz[[1, 0]] - 1) / 2, self.pos[[1, 0]] + (self.target_sz[[1, 0]] + 1) / 2))
        new_state = new_state.clamp(min=1, max=torch.max(self.image_sz))
        new_state[2:] = new_state[2:] - new_state[:2]
        ###
        # print("new_state: {}".format(new_state))

        # use trseg to generate the fined mask based on the detected center and box
        Cbbox = new_state.clone().tolist()
        Cframe = image
        if not self.valid_box(Cbbox):
            box = self.init_state.copy()
            box[0], box[1] = self.pos[1].item(), self.pos[0].item()
            Cbbox = box

        mask_update_step = 16
        update_mask_target = False
        if (flag not in ['not_found', 'uncertain']) and max_score > self.max_score_mask:
            self.mask_update_Cframe = copy.deepcopy(Cframe)
            self.mask_update_bbox = None
            self.max_score_mask = max_score

        if self.frame_num - self.last_update_mask > mask_update_step and self.mask_update_Cframe is not None:
            self.last_update_mask = self.frame_num
            update_mask_target = True

        pred_mask, pred_bbox = self.trseg.forward_mask(Cframe, np.array(Cbbox), update_mask_crop=update_mask_target, update_data=(self.mask_update_Cframe, self.mask_update_bbox))  # [H, W]
        if update_mask_target:
            self.mask_update_Cframe = None
            self.mask_update_bbox = None
            self.max_score_mask = -100
        elif self.mask_update_Cframe is not None and self.mask_update_bbox is None:
            self.mask_update_bbox = copy.deepcopy(pred_bbox)
        # print("pred_bbox: {}".format(pred_bbox))
        # pred_bbox: x0y0wh
        THRES = getattr(self.params, 'THRES', 0.50)
        final_mask = (pred_mask > THRES).astype(np.uint8)

        num_mask = np.sum(final_mask)
        if num_mask <= 4:
            final_mask = self.prev_mask
        else:
            self.prev_mask = final_mask
        update_cls_using_mask = getattr(self.params, 'update_cls_using_mask', True)
        # update_cls_using_mask = True
        if update_cls_using_mask and flag != 'not_found' and self.valid_box(pred_bbox):
            mask_bbox = rect_from_mask(final_mask)  # numpy array, [x0,y0,w,h]
            # print("pred_bbox_from_mask: {}".format(mask_bbox))
            new_pos, new_target_sz = self.convert_bbox(pred_bbox)
            new_pos, new_target_sz = new_pos.to(self.prev_pos.device), new_target_sz.to(self.target_scale.device)

            scale_lr = 0.15
            new_target_sz = scale_lr * new_target_sz + (1-scale_lr) * self.target_sz

            new_scale = torch.sqrt(new_target_sz.prod() / self.base_target_sz.prod())
            self.target_scale = new_scale.clamp(self.min_scale_factor, self.max_scale_factor)
            # self.prev_pos = new_pos

            self.prev_pos = new_pos

            self.pos = new_pos
            self.target_sz = new_target_sz

        # Compute output bounding box
        new_state = torch.cat(
            (self.pos[[1, 0]] - (self.target_sz[[1, 0]] - 1) / 2, self.target_sz[[1, 0]]))  # x, y, w, h
        new_state = new_state.clamp(min=1, max=torch.max(self.image_sz))
        # used to compute the iou between current predict target box and the previous predicted target box.
        self.prev_box = torch.cat(
            (self.pos[[0, 1]] - (self.target_sz[[0, 1]] - 1) / 2, self.pos[[0, 1]] + (self.target_sz[[0, 1]] - 1) / 2))

        ### vis mask
        # hm = final_mask  # [h, w]
        # hm = scores_vis.squeeze().cpu().numpy()
        # save_path = '/home/cyt/data/experiments/vot2021_vis/mask/'
        # if self.frame_num == 2:
        #     for i in range(1, 1000):
        #             self.seq_path = save_path + str(i)
        #             if not os.path.exists(self.seq_path):
        #                 os.makedirs(self.seq_path)
        #                 break
        # save_mask_dir = os.path.join('{}/{}.png'.format(self.seq_path, self.frame_num))
        # hm = np.uint8(255 * hm)
        # hm = cv2.applyColorMap(hm, cv2.COLORMAP_JET)
        # cv2.imwrite(save_mask_dir, hm)
        ### vis end


        ### vis track
        if vis_switch:
            img_patch = self.im_patches[scale_ind, ...].clone().squeeze().cpu().permute(1, 2, 0).numpy()
            center_288 = (max_disp * 16).cpu().int().numpy() # ycxc
            second_center_288 = (max_disp2 * 16).cpu().int().numpy()
            corner_bbox = torch.tensor([pred_bbox[1], pred_bbox[0], pred_bbox[3], pred_bbox[2]]) # y0x0hw
            corner_bbox[2:] = corner_bbox[:2] + corner_bbox[2:] # y0x0y1x1
            corner_bbox[:2] = (corner_bbox[:2] - sample_pos[scale_ind, :]) / (sample_scales[scale_ind] + 1e-6) / (self.img_support_sz / test_feat_3.size(-1) + 1e-6)  # translation_vec
            corner_bbox[2:] = (corner_bbox[2:] - sample_pos[scale_ind, :]) / (sample_scales[scale_ind] + 1e-6) / (self.img_support_sz / test_feat_3.size(-1) + 1e-6)  # translation_vec
            corner_bbox = corner_bbox + 9
            corner_bbox_288 = (corner_bbox * 16).cpu().int().numpy() # y0x0y1x1
            corner_bbox_288 = np.array([corner_bbox_288[1], corner_bbox_288[0], corner_bbox_288[3], corner_bbox_288[2]]) # x0y0x1y1
            if flag == 'normal':
                vis_track = [img_patch, center_288, corner_bbox_288, 1]
            elif flag == 'hard_negative':
                vis_track = [img_patch, center_288, corner_bbox_288, 2, second_center_288]
            elif flag == 'not_found':
                vis_track = [img_patch, center_288, corner_bbox_288, 0]
            else: #  flag == 'uncertain'
                vis_track = [img_patch, center_288, corner_bbox_288, -1]
        else:
            vis_track = None


        ###
        # ----------------- Online Traing ------------------ #  后面必须把这部分抽离出去，不然看着太杂乱

        update_flag = flag not in ['not_found', 'uncertain']
        hard_negative = (flag == 'hard_negative')
        learning_rate = getattr(self.params, 'hard_negative_learning_rate', None) if hard_negative else None

        if getattr(self.params, 'update_classifier_and_regressor', False) and update_flag:
            # Get train sample
            train_x_3 = test_feat_3[scale_ind:scale_ind + 1, ...].clone()
            train_x_23 = test_feat_23[scale_ind:scale_ind + 1, ...].clone()
            train_x_123 = test_feat_123[scale_ind:scale_ind + 1, ...].clone()

            # Create target_box and label for spatial sample
            target_box = self.get_training_box(self.pos, self.target_sz, sample_pos[scale_ind, :], sample_scales[scale_ind])

            if hard_negative:
            # if hard_negative and max_score > 0.1: #########################################
                self.update_classifier(train_x_3, train_x_23, train_x_123, target_box, learning_rate, s[scale_ind, ...])
                # self.update_transcls(train_x_transCls, target_box)
            if (self.frame_num - 1) <= self.params.init_train_frames:
                self.update_classifier(train_x_3, train_x_23, train_x_123, target_box)
                # self.update_transcls(train_x_transCls, target_box)
            elif (self.frame_num - 1) > self.params.init_train_frames:
                if max_score >= self.max_score or self.train_x_3 is None:
                    self.max_score = max_score
                    # self.train_x_transCls = train_x_transCls
                    self.target_box = target_box
                    self.train_x_3 = train_x_3
                    self.train_x_23 = train_x_23
                    self.train_x_123 = train_x_123
                    self.s = s[scale_ind, ...]
                if (self.frame_num - 1) % self.params.train_skipping == 0:
                    if self.params.ues_select_sample_strategy:
                        # if self.max_score > 0.1: #########################################
                        self.update_classifier(self.train_x_3, self.train_x_23, self.train_x_123, self.target_box)
                        # self.update_transcls(self.train_x_transCls, self.target_box)
                    else:
                        # if max_score > 0.1: #########################################
                        self.update_classifier(train_x_3, train_x_23, train_x_123, target_box)
                        # self.update_transcls(train_x_transCls, target_box)
                    self.train_x_3 = None
                    self.max_score = -100

        # ----------------- Online Traing End ------------------ #

        out = {'target_bbox': new_state.tolist(), 'score_map_18': s[scale_ind, ...], 'score_map_72': score_map,
               'image': self.im_patches[scale_ind, ...].squeeze(), 'mask': final_mask, 'vis': vis_track, 'max_score': max_score}
        return out

    def valid_box(self, bbox):
        """
        :param bbox: [x,y,w,h]
        :return:
        """
        if bbox[2] <= 1 or bbox[3] <= 1:
            return False
        '''TODO: 需要添加更多的判定条件'''
        return True

    def tensors_gray2rgb(self, tensor, image):
        """Convert to rgb images [tenosrs： (num, 1, h, w)].
           :return tensors： (num, 3, h, w)
        """
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

        save_path = '/home/cyt/data/experiments/treg_feat_vis/after_treg_reg_feat/'
        if self.frame_num == 2:
            for i in range(1, 1000):
                self.seq_path = save_path + str(i)
                if not os.path.exists(self.seq_path):
                    os.makedirs(self.seq_path)
                    break
        # if not os.path.exists('/home/cyt/data/experiments/treg_feat_vis/before_treg_reg_feat/{}'.format(self.seq_id)):
        #     os.makedirs('/home/cyt/data/experiments/treg_feat_vis/before_treg_reg_feat/{}'.format(self.seq_id))
        # tmp = os.path.join('/home/cyt/data/experiments/treg_feat_vis/before_treg_reg_feat/{}/{}.png'.format(self.seq_id,
        #                                                                                                     self.frame_num))
        # if not os.path.exists('/home/cyt/data/experiments/treg_feat_vis/after_treg_reg_feat/{}'.format(self.seq_id)):
        #     os.makedirs('/home/cyt/data/experiments/treg_feat_vis/after_treg_reg_feat/{}'.format(self.seq_id))
        tmp = os.path.join('{}/{}.png'.format(self.seq_path, self.frame_num))

        hm = tensor.clone().squeeze().cpu().detach().numpy()  # [h, w]

        #
        # min_ = np.min(hm[20:50, 20:50])
        # hm = np.maximum(hm, 0)
        # hm = np.minimum(hm, 0)
        # hm = hm + abs(min_)
        hm = np.fabs(hm)
        # hm = np.absolute(hm)
        max_ = np.max(hm[20:50, 20:50])
        # print("max_v : {}".format(max_))
        hm = np.minimum(hm, max_)
        # hm /= max_
        f_max = np.max(hm)
        f_min = np.min(hm)
        hm = (hm - f_min) / (f_max - f_min)
        hm = cv2.resize(hm, (image.shape[1], image.shape[0]))
        hm = np.uint8(255 * hm)

        hm = cv2.applyColorMap(hm, cv2.COLORMAP_JET)

        img = hm * 0.4 + image

        cv2.imwrite(tmp, img)

        return img

    def save_pred_image(self, img, bbox, center, center2, save_file=None):
        """Display a 2D tensor.
        """
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        draw_rec = cv2.rectangle(img, (bbox[1], bbox[0]), (bbox[3], bbox[2]), (0, 0, 255), 2)
        draw_rec = cv2.circle(draw_rec, (center[1], center[0]), 3, (0, 0, 255), -1)
        draw_rec = cv2.circle(draw_rec, (center2[1], center2[0]), 3, (0, 255, 0), -1)
        cv2.imwrite(save_file, draw_rec)
        img = cv2.cvtColor(draw_rec, cv2.COLOR_BGR2RGB)
        img_tensor = torch.from_numpy(img).float().permute(2, 0, 1)
        return img_tensor

    def iou_pred(self, pred, target):
        pred_left = pred[1]
        pred_top = pred[0]
        pred_right = pred[3]
        pred_bottom = pred[2]

        target_left = target[1]
        target_top = target[0]
        target_right = target[3]
        target_bottom = target[2]

        target_area = (target_right - target_left) * \
                      (target_bottom - target_top)
        pred_area = (pred_right - pred_left) * \
                    (pred_bottom - pred_top)

        w_intersect = torch.min(pred_right, target_right) - \
                      torch.max(pred_left, target_left)
        h_intersect = torch.min(pred_bottom, target_bottom) - \
                      torch.max(pred_top, target_top)

        area_intersect = w_intersect * h_intersect
        area_union = target_area + pred_area - area_intersect

        iou = area_intersect / area_union

        return iou.item()

    def ctdet_decode(self, w2h2, max_disp):
        max_disp = max_disp.long()
        w2h2 = w2h2.squeeze()
        w2h2 = (torch.relu(w2h2)) * 4.0
        w2h2_max = w2h2[..., max_disp[0], max_disp[1]]
        # w2h2_max = w2h2[..., max_disp[1], max_disp[0]]
        return w2h2_max

    def get_sample_location(self, sample_coord):
        """Get the location of the extracted sample."""
        sample_coord = sample_coord.float()
        sample_pos = 0.5 * (sample_coord[:, :2] + sample_coord[:, 2:] - 1)
        sample_scales = ((sample_coord[:, 2:] - sample_coord[:, :2]) / self.img_sample_sz).prod(dim=1).sqrt()
        return sample_pos, sample_scales

    def get_centered_sample_pos(self, feature_sz, kernel_size):
        """Get the center position for the new sample. Make sure the target is correctly centered."""
        return self.pos + ((feature_sz + kernel_size) % 2) * self.target_scale * \
               self.img_support_sz / (2 * feature_sz)

    def classify_target_3(self, sample_x: TensorList):
        """Classify target by applying the FCOT classfication filter with size of 18.
        """
        with torch.no_grad():
            scores_cur = self.net.classifier_3.classify(self.target_filter_3, sample_x)
            scores_init = self.net.classifier_3.classify(self.init_target_filter_3, sample_x)
        scores = (1.0 - self.params.lamda_18) * scores_init + self.params.lamda_18 * scores_cur
        return scores

    def classify_target_23(self, sample_x: TensorList):
        """Classify target by applying the FCOT classfication filter with size of 18.
        """
        with torch.no_grad():
            scores_cur = self.net.classifier_2.classify(self.target_filter_23, sample_x)
            scores_init = self.net.classifier_2.classify(self.init_target_filter_23, sample_x)
        scores = (1.0 - self.params.lamda_18) * scores_init + self.params.lamda_18 * scores_cur
        return scores

    def classify_target_123(self, sample_x: TensorList):
        """Classify target by applying the FCOT classfication filter with size of 18.
        """
        with torch.no_grad():
            scores_cur = self.net.classifier_1.classify(self.target_filter_123, sample_x)
            scores_init = self.net.classifier_1.classify(self.init_target_filter_123, sample_x)
        scores = (1.0 - self.params.lamda_18) * scores_init + self.params.lamda_18 * scores_cur
        return scores

    def localize_target(self, scores, sample_scales, feature_sz):
        """Run the target localization."""

        scores = scores.squeeze(1)

        if getattr(self.params, 'advanced_localization', False):
            return self.localize_advanced(scores, sample_scales, feature_sz)

        # Get maximum
        score_sz = torch.Tensor(list(scores.shape[-2:]))
        score_center = (score_sz - 1) / 2
        max_score, max_disp = dcf.max2d(scores)
        _, scale_ind = torch.max(max_score, dim=0)
        max_disp = max_disp[scale_ind, ...].float().cpu().view(-1)
        target_disp = max_disp - score_center

        # Compute translation vector and scale change factor
        translation_vec = target_disp * (self.img_support_sz / feature_sz) * sample_scales[scale_ind]

        return translation_vec, scale_ind, scores, None

    def localize_advanced(self, scores, sample_scales, feature_sz):
        """Run the target advanced localization (as in ATOM).
            w2h2 shape: (1, 1, 4, 72, 72)
        """
        if scores.dim() == 4:
            scores.squeeze(1)

        sz = scores.shape[-2:]
        score_sz = torch.Tensor(list(sz))
        score_center = (score_sz - 1) / 2

        scores_hn = scores
        if self.output_window is not None and getattr(self.params, 'perform_hn_without_windowing', False):
            scores_hn = scores.clone()
            scores *= self.output_window

        max_score1, max_disp1 = dcf.max2d(scores)
        _, scale_ind = torch.max(max_score1, dim=0)
        sample_scale = sample_scales[scale_ind]
        max_score1 = max_score1[scale_ind]
        max_disp1 = max_disp1[scale_ind, ...].float().cpu().view(-1)
        target_disp1 = max_disp1 - score_center
        translation_vec1 = target_disp1 * (self.img_support_sz / feature_sz) * sample_scale

        max_score1_hn, max_disp1_hn = dcf.max2d(scores_hn)
        max_score1_hn = max_score1_hn[scale_ind]
        print("max_score: {}, max_score_hn: {}".format(max_score1.item(), max_score1_hn.item()))

        # Mask out target neighborhood
        target_neigh_sz = self.params.target_neighborhood_scale * (self.target_sz / sample_scale) * (
                feature_sz / self.img_support_sz)

        tneigh_top = max(round(max_disp1[0].item() - target_neigh_sz[0].item() / 2), 0)
        tneigh_bottom = min(round(max_disp1[0].item() + target_neigh_sz[0].item() / 2 + 1), sz[0])
        tneigh_left = max(round(max_disp1[1].item() - target_neigh_sz[1].item() / 2), 0)
        tneigh_right = min(round(max_disp1[1].item() + target_neigh_sz[1].item() / 2 + 1), sz[1])
        scores_masked = scores_hn[scale_ind:scale_ind + 1, ...].clone()
        scores_masked[..., tneigh_top:tneigh_bottom, tneigh_left:tneigh_right] = 0

        # Find new maximum
        max_score2, max_disp2 = dcf.max2d(scores_masked)
        # print("max_score2: {}".format(max_score2.item()))

        max_disp2 = max_disp2.float().cpu().view(-1)
        target_disp2 = max_disp2 - score_center
        translation_vec2 = target_disp2 * (self.img_support_sz / feature_sz) * sample_scale

        if max_score1.item() < self.params.target_not_found_threshold:
            return translation_vec1, scale_ind, scores_hn, 'not_found', max_disp1, max_disp2, None

        # Handle the different cases
        if max_score2 > self.params.distractor_threshold * max_score1:
            disp_norm1 = torch.sqrt(torch.sum(target_disp1 ** 2))
            disp_norm2 = torch.sqrt(torch.sum(target_disp2 ** 2))
            disp_threshold = self.params.dispalcement_scale * math.sqrt(sz[0] * sz[1]) / 2

            if disp_norm2 > disp_threshold and disp_norm1 < disp_threshold:
                return translation_vec1, scale_ind, scores_hn, 'hard_negative', max_disp1, max_disp2, max_score1.item()
            if disp_norm2 < disp_threshold and disp_norm1 > disp_threshold:
                return translation_vec2, scale_ind, scores_hn, 'hard_negative', max_disp2, max_disp1, max_score2.item()
            if disp_norm2 > disp_threshold and disp_norm1 > disp_threshold:
                return translation_vec1, scale_ind, scores_hn, 'uncertain', max_disp1, max_disp2, max_score1.item()

            # If also the distractor is close, return with highest score
            return translation_vec1, scale_ind, scores_hn, 'uncertain', max_disp1, max_disp2, max_score1.item()

        if max_score1_hn > 0.15 and max_score2 > self.params.hard_negative_threshold * max_score1 and max_score2 > self.params.target_not_found_threshold:
            return translation_vec1, scale_ind, scores_hn, 'hard_negative', max_disp1, max_disp2, max_score1.item()

        return translation_vec1, scale_ind, scores_hn, 'normal', max_disp1, max_disp2, max_score1.item()

    def extract_backbone_features(self, im: torch.Tensor, pos: torch.Tensor, scales, sz: torch.Tensor):
        # im_patches, patch_coords = sample_patch_multiscale(im, pos, scales, sz, getattr(self.params, 'border_mode', 'replicate'))
        im_patches, patch_coords = sample_patch_multiscale(im, pos, scales, sz,
                                                           mode=getattr(self.params, 'border_mode', 'replicate'),
                                                           max_scale_change=getattr(self.params,
                                                                                    'patch_max_scale_change',
                                                                                    None))
        with torch.no_grad():
            backbone_feat = self.net.extract_backbone(im_patches.clone())
            # transcls_backbone_feat = self.transCls_net.extract_backbone(im_patches.clone())
        self.im_patches = im_patches
        return backbone_feat, patch_coords

    def get_classification_features(self, backbone_feat):
        with torch.no_grad():
            return self.net.extract_classification_feat_18(backbone_feat)

    def generate_init_samples(self, im: torch.Tensor): # -> TensorList:
        """Perform data augmentation to generate initial training samples."""

        if getattr(self.params, 'border_mode', 'replicate') == 'inside':
            # Get new sample size if forced inside the image
            im_sz = torch.Tensor([im.shape[2], im.shape[3]])
            sample_sz = self.target_scale * self.img_sample_sz
            shrink_factor = (sample_sz.float() / im_sz).max().clamp(1)
            sample_sz = (sample_sz.float() / shrink_factor)
            self.init_sample_scale = (sample_sz / self.img_sample_sz).prod().sqrt()
            tl = self.pos - (sample_sz - 1) / 2
            br = self.pos + sample_sz / 2 + 1
            global_shift = - ((-tl).clamp(0) - (br - im_sz).clamp(0)) / self.init_sample_scale
        else:
            self.init_sample_scale = self.target_scale
            global_shift = torch.zeros(2)

        self.init_sample_pos = self.pos.round()

        # Compute augmentation size
        aug_expansion_factor = getattr(self.params, 'augmentation_expansion_factor', None)  # 2
        aug_expansion_sz = self.img_sample_sz.clone()  # [288, 288]
        aug_output_sz = None
        if aug_expansion_factor is not None and aug_expansion_factor != 1:
            aug_expansion_sz = (self.img_sample_sz * aug_expansion_factor).long()
            aug_expansion_sz += (aug_expansion_sz - self.img_sample_sz.long()) % 2
            aug_expansion_sz = aug_expansion_sz.float()
            aug_output_sz = self.img_sample_sz.long().tolist()

        # Random shift for each sample
        get_rand_shift = lambda: None
        random_shift_factor = getattr(self.params, 'random_shift_factor', 0)  # 1/3
        if random_shift_factor > 0:
            get_rand_shift = lambda: (
                        (torch.rand(2) - 0.5) * self.img_sample_sz * random_shift_factor + global_shift).long().tolist()

        # Always put identity transformation first, since it is the unaugmented sample that is always used
        self.transforms = [augmentation.Identity(aug_output_sz, global_shift.long().tolist())]

        augs = self.params.augmentation if getattr(self.params, 'use_augmentation', True) else {}

        # Add all augmentations
        if 'scale' in augs:
            self.transforms.extend(
                [augmentation.Scale(scale_factor, aug_output_sz, get_rand_shift()) for scale_factor in augs['scale']])
        if 'shift' in augs:
            self.transforms.extend(
                [augmentation.Translation(shift, aug_output_sz, global_shift.long().tolist()) for shift in
                 augs['shift']])
        if 'relativeshift' in augs:
            get_absolute = lambda shift: (torch.Tensor(shift) * self.img_sample_sz / 2).long().tolist()
            self.transforms.extend(
                [augmentation.Translation(get_absolute(shift), aug_output_sz, global_shift.long().tolist()) for shift in
                 augs['relativeshift']])
        if 'fliplr' in augs and augs['fliplr']:
            self.transforms.append(augmentation.FlipHorizontal(aug_output_sz, get_rand_shift()))
        if 'blur' in augs:
            self.transforms.extend(
                [augmentation.Blur(sigma, aug_output_sz, get_rand_shift()) for sigma in augs['blur']])
        if 'rotate' in augs:
            self.transforms.extend(
                [augmentation.Rotate(angle, aug_output_sz, get_rand_shift()) for angle in augs['rotate']])

        # Extract augmented image patches
        im_patches = sample_patch_transformed(im, self.init_sample_pos, self.init_sample_scale, aug_expansion_sz,
                                              self.transforms)

        # Extract initial backbone features
        with torch.no_grad():
            init_backbone_feat = self.net.extract_backbone(im_patches.clone())
            # init_transcls_backbone_feat = self.transCls_net.extract_backbone(im_patches[:self.params.transcls_mem_size].clone())

        return init_backbone_feat

    def init_target_boxes(self):
        """Get the target bounding boxes for the initial augmented samples."""
        self.classifier_target_box = self.get_training_box(self.pos, self.target_sz, self.init_sample_pos,
                                                           self.init_sample_scale)
        init_target_boxes = TensorList()
        for T in self.transforms:
            init_target_boxes.append(self.classifier_target_box + torch.Tensor([T.shift[1], T.shift[0], 0, 0]))
        init_target_boxes = torch.cat(init_target_boxes.view(1, 4), 0).to(self.params.device)
        self.target_boxes = init_target_boxes.new_zeros(self.params.sample_memory_size, 4)
        self.target_boxes[:init_target_boxes.shape[0], :] = init_target_boxes
        self.target_boxes_cls = self.target_boxes.clone()
        # print("target_boxes shape: {}".format(self.target_boxes_reg_72.shape))
        return init_target_boxes

    def init_memory(self, train_x_3: TensorList, train_x_23: TensorList, train_x_123: TensorList):
        # Initialize first-frame spatial training samples
        self.num_init_samples = train_x_3.size(0)
        init_sample_weights = TensorList([x.new_ones(1) / x.shape[0] for x in train_x_3])

        # Sample counters and weights for spatial
        # self.num_stored_samples = self.num_init_samples.copy()
        # self.previous_replace_ind = [None] * len(self.num_stored_samples)
        # self.sample_weights = TensorList([x.new_zeros(self.params.sample_memory_size) for x in train_x_72_cls])
        # for sw, init_sw, num in zip(self.sample_weights, init_sample_weights, self.num_init_samples):
        #     sw[:num] = init_sw
        # print(self.sample_weights)

        self.num_stored_samples_cls = self.num_init_samples.copy()
        self.previous_replace_ind_cls = [None] * len(self.num_stored_samples_cls)
        self.sample_weights_cls = TensorList([x.new_zeros(self.params.sample_memory_size) for x in train_x_3])
        for sw, init_sw, num in zip(self.sample_weights_cls, init_sample_weights, self.num_init_samples):
            sw[:num] = init_sw

        # initialize classification features memory
        self.training_samples_cls_3 = TensorList([x.new_zeros(self.params.sample_memory_size, x.shape[1], x.shape[2], x.shape[3]) for x in train_x_3])
        for ts, x in zip(self.training_samples_cls_3, train_x_3):
            ts[:x.shape[0], ...] = x

        self.training_samples_cls_23 = TensorList(
            [x.new_zeros(self.params.sample_memory_size, x.shape[1], x.shape[2], x.shape[3]) for x in train_x_23])
        for ts, x in zip(self.training_samples_cls_23, train_x_23):
            ts[:x.shape[0], ...] = x

        self.training_samples_cls_123 = TensorList(
            [x.new_zeros(self.params.sample_memory_size, x.shape[1], x.shape[2], x.shape[3]) for x in train_x_123])
        for ts, x in zip(self.training_samples_cls_123, train_x_123):
            ts[:x.shape[0], ...] = x


    def update_memory_cls(self, sample_x_cls_3: TensorList, sample_x_cls_23: TensorList, sample_x_cls_123: TensorList,
                          target_box, learning_rate=None):
        # Update weights and get replace ind
        replace_ind = self.update_sample_weights(self.sample_weights_cls, self.previous_replace_ind_cls,
                                                 self.num_stored_samples_cls, self.num_init_samples, learning_rate)
        self.previous_replace_ind_cls = replace_ind
        # print("replace: {}".format(replace_ind))
        # print(self.sample_weights)

        # Update sample and label memory
        for train_samp, x, ind in zip(self.training_samples_cls_3, sample_x_cls_3, replace_ind):
            train_samp[ind:ind + 1, ...] = x
        for train_samp, x, ind in zip(self.training_samples_cls_23, sample_x_cls_23, replace_ind):
            train_samp[ind:ind + 1, ...] = x
        for train_samp, x, ind in zip(self.training_samples_cls_123, sample_x_cls_123, replace_ind):
            train_samp[ind:ind + 1, ...] = x

        # Update bb memory
        self.target_boxes_cls[replace_ind[0], :] = target_box

        self.num_stored_samples_cls += 1


    def update_transcls(self, sample_x_transcls, target_box):
        # Update weights and get replace ind
        replace_ind = [self.previous_replace_ind_transcls]
        self.previous_replace_ind_transcls += 1
        if self.previous_replace_ind_transcls == int(self.params.transcls_mem_size):
            self.previous_replace_ind_transcls = 1

        with torch.no_grad():
            target_cur = sample_x_transcls.view(1, *sample_x_transcls.shape[-3:])
            bb_pool = target_box.clone().to(self.params.device)
            target_cur = self.transCls_net.target_pool(target_cur, bb_pool)  # 1, C, Ht, Wt
            target_cur = target_cur.view(1, -1, 4, 4)
            target_cur = target_cur.flatten(2).permute(2, 0, 1)  #
            # print("target_cur shape: {}".format(target_cur.shape))
            target_cur = self.transCls_net.transformer.encoder2(target_cur)
            target_cur = self.transCls_net.transformer.box_head1(target_cur.permute(2, 1, 0)).permute(2, 1,0)  # 1, 1, C
            target_cur = target_cur.view(1, -1)
            self.transCls_target[replace_ind] = target_cur  # [3, 1, C]


    def update_sample_weights(self, sample_weights, previous_replace_ind, num_stored_samples, num_init_samples,
                              learning_rate=None):
        # Update weights and get index to replace
        replace_ind = []
        for sw, prev_ind, num_samp, num_init in zip(sample_weights, previous_replace_ind, num_stored_samples,
                                                    num_init_samples):
            lr = learning_rate
            if lr is None:
                lr = self.params.learning_rate

            init_samp_weight = getattr(self.params, 'init_samples_minimum_weight', None)
            if init_samp_weight == 0:
                init_samp_weight = None
            s_ind = 0 if init_samp_weight is None else num_init

            if num_samp == 0 or lr == 1:
                sw[:] = 0
                sw[0] = 1
                r_ind = 0
            else:
                # Get index to replace
                if num_samp < sw.shape[0]:
                    r_ind = num_samp
                else:
                    _, r_ind = torch.min(sw[s_ind:], 0)
                    r_ind = r_ind.item() + s_ind

                # Update weights
                if prev_ind is None:
                    sw /= 1 - lr
                    sw[r_ind] = lr
                else:
                    sw[r_ind] = sw[prev_ind] / (1 - lr)

            sw /= sw.sum()
            if init_samp_weight is not None and sw[:num_init].sum() < init_samp_weight:
                sw /= init_samp_weight + sw[num_init:].sum()
                sw[:num_init] = init_samp_weight / num_init

            replace_ind.append(r_ind)

        return replace_ind

    def get_training_box(self, pos, sz, sample_pos, sample_scale):
        """All inputs in original image coordinates.
        Generates a box in the cropped image sample reference frame, in the format used by the IoUNet."""
        box_center = (pos - sample_pos) / sample_scale + (self.img_sample_sz - 1) / 2
        box_sz = sz / sample_scale
        target_ul = box_center - (box_sz - 1) / 2
        return torch.cat([target_ul.flip((0,)), box_sz.flip((0,))])

    def init_classifier(self, init_backbone_feat):
        print("init classifiers")
        # Get classification features
        x = self.net.get_backbone_clf_feat(init_backbone_feat)
        # train_feat_18_cls = self.get_classification_features(init_backbone_feat)

        with torch.no_grad():
            train_feat_3, train_feat_23, train_feat_123 = self.net.extract_3scales_feat(init_backbone_feat)
            train_feat_3 = self.net.classifier_3.extract_classification_feat(train_feat_3.view(-1, *train_feat_3.shape[-3:]))
            train_feat_23 = self.net.classifier_2.extract_classification_feat(train_feat_23.view(-1, *train_feat_23.shape[-3:]))
            train_feat_123 = self.net.classifier_1.extract_classification_feat(train_feat_123.view(-1, *train_feat_123.shape[-3:]))
            # print(train_feat_3.shape, train_feat_23.shape, train_feat_123.shape)

        # Add the dropout augmentation here, since it requires extraction of the classification features
        if 'dropout' in self.params.augmentation and getattr(self.params, 'use_augmentation', True):
            num, prob = self.params.augmentation['dropout']
            self.transforms.extend(self.transforms[:1] * num)
            train_feat_3 = torch.cat([train_feat_3, F.dropout2d(train_feat_3[0:1, ...].expand(num, -1, -1, -1), p=prob, training=True)])
            train_feat_23 = torch.cat([train_feat_23, F.dropout2d(train_feat_23[0:1, ...].expand(num, -1, -1, -1), p=prob, training=True)])
            train_feat_123 = torch.cat([train_feat_123, F.dropout2d(train_feat_123[0:1, ...].expand(num, -1, -1, -1), p=prob, training=True)])

        # Get target boxes for the different augmentations
        target_boxes = self.init_target_boxes()
        # target_boxes = self.target_boxes.clone()

        # Set number of iterations
        num_iter = getattr(self.params, 'net_opt_iter', None)
        # num_iter_72 = getattr(self.params, 'net_opt_iter_72', None)

        # Get target filter by running the discriminative model prediction module
        with torch.no_grad():
            # extract target_filter_72, target_filter_18 and target_reg_filter_72 using Clf and Reg model generators.
            self.target_filter_3, _, _ = self.net.classifier_3.get_filter(train_feat_3, target_boxes, num_iter=num_iter)
            self.target_filter_23, _, _ = self.net.classifier_2.get_filter(train_feat_23, target_boxes, num_iter=num_iter)
            self.target_filter_123, _, _ = self.net.classifier_1.get_filter(train_feat_123, target_boxes, num_iter=num_iter)

            # get initial Clf and Reg model used in tracking process, which merge the initial model and the optimized model.
            self.init_target_filter_3 = self.target_filter_3.clone()
            self.init_target_filter_23 = self.target_filter_23.clone()
            self.init_target_filter_123 = self.target_filter_123.clone()

        # Set feature size and other related sizes
        self.feature_sz_18 = torch.Tensor(list(x.shape[-2:]))
        ksz_18 = self.net.classifier_3.filter_size
        self.kernel_size_18 = torch.Tensor([ksz_18, ksz_18] if isinstance(ksz_18, (int, float)) else ksz_18)
        self.output_sz = self.feature_sz_18 + (self.kernel_size_18 + 1) % 2
        # self.output_sz = torch.Tensor([18, 18])  # 5.12加，调一下transCls分支

        # Construct output window
        self.output_window = None
        if getattr(self.params, 'window_output', False):
            if getattr(self.params, 'use_clipped_window', False):
                self.output_window = dcf.hann2d_clipped(
                    self.output_sz.long(),
                    self.output_sz.long() * self.params.effective_search_area / self.params.search_area_scale,
                    centered=False).to(self.params.device)
            else:
                self.output_window = dcf.hann2d(self.output_sz.long(), centered=True).to(self.params.device)
            
            # self.output_window = self.output_window.squeeze(0)
            smooth_weight = 0.1
            self.output_window = self.output_window.squeeze(0) * (1-smooth_weight) + smooth_weight

        # Init memory
        if getattr(self.params, 'update_classifier_and_regressor', True):
            self.init_memory(TensorList([train_feat_3]), TensorList([train_feat_23]), TensorList([train_feat_123]))

    def update_classifier(self, train_x_cls_3, train_x_cls_23, train_x_cls_123, target_box, learning_rate=None, scores=None):
        # Set flags and learning rate
        hard_negative_flag = learning_rate is not None
        if learning_rate is None:
            learning_rate = self.params.learning_rate

        # Update the tracker memory
        self.update_memory_cls(TensorList([train_x_cls_3]), TensorList([train_x_cls_23]), TensorList([train_x_cls_123]),
                               target_box, learning_rate)

        # Decide the number of iterations to run
        num_iter = 0
        low_score_th = getattr(self.params, 'low_score_opt_threshold', None)
        if hard_negative_flag:
            num_iter = getattr(self.params, 'net_opt_hn_iter', None)
        elif low_score_th is not None and low_score_th > scores.max().item():
            num_iter = getattr(self.params, 'net_opt_low_iter', None)
        elif (self.frame_num - 1) <= self.params.init_train_frames:
            num_iter = self.params.init_train_iter
        elif (self.frame_num - 1) % self.params.train_skipping == 0:
            num_iter = getattr(self.params, 'net_opt_update_iter', None)

        if num_iter > 0:
            # Get inputs for the DiMP filter optimizer module
            samples_cls_3 = self.training_samples_cls_3[0][:self.num_stored_samples_cls[0], ...]
            samples_cls_23 = self.training_samples_cls_23[0][:self.num_stored_samples_cls[0], ...]
            samples_cls_123 = self.training_samples_cls_123[0][:self.num_stored_samples_cls[0], ...]
            target_boxes = self.target_boxes_cls[:self.num_stored_samples_cls[0], :]
            sample_weights = self.sample_weights_cls[0][:self.num_stored_samples_cls[0]]

            # Run the model optimizer module
            with torch.no_grad():
                self.target_filter_3, _, _ = self.net.classifier_3.filter_optimizer(self.target_filter_3,
                                                                                    samples_cls_3,
                                                                                    target_boxes.clone(),
                                                                                    sample_weight=sample_weights,
                                                                                    num_iter=num_iter)
                self.target_filter_23, _, _ = self.net.classifier_2.filter_optimizer(self.target_filter_23,
                                                                                    samples_cls_23,
                                                                                    target_boxes.clone(),
                                                                                    sample_weight=sample_weights,
                                                                                    num_iter=num_iter)
                self.target_filter_123, _, _ = self.net.classifier_1.filter_optimizer(self.target_filter_123,
                                                                                    samples_cls_123,
                                                                                    target_boxes.clone(),
                                                                                    sample_weight=sample_weights,
                                                                                    num_iter=num_iter)
